//	sdsf_shared.c
//	SDSF
//
//	Created by Nils Daumann on 25.06.11.
//	Copyright (c) 2011 Nils Daumann (www.slindev.com)

//	Permission is hereby granted, free of charge, to any person obtaining a copy
//	of this software and associated documentation files (the "Software"), to deal
//	in the Software without restriction, including without limitation the rights
//	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//	copies of the Software, and to permit persons to whom the Software is
//	furnished to do so, subject to the following conditions:

//	The above copyright notice and this permission notice shall be included in
//	all copies or substantial portions of the Software.

//	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//	THE SOFTWARE.

#ifndef __SDSF_SHARED_C__
#define __SDSF_SHARED_C__

#ifndef SDSF_FILEMAPPING_FILE
#define SDSF_FILEMAPPING_FILE "slindev/shaders/sdsf_filemapping.fm"
#endif

/**
 * Precompile fx file.
 * Precompiles the given .fx (has to be called .fxo) file to a .fxo file with defines and stores the name id into a file.
 * @param fxfile name of the .fx file to precompile with .fxo as extension.
 * @param defines a string with the defines to compile the shader with, seperated by spaces: "TESTDEF=1 ANOTHERDEF=1".
 */
void sdsf_fx_compile(STRING *fxfile, STRING *defines)
{
	STRING *fn = str_create(fxfile);
	if(str_stri(fxfile, ".fxo"))
	{
		str_trunc(fn, 1);
		if(file_exists(fn))
		{
			STRING *params = str_create("/T fx_2_0 ");
			STRING *outfile = str_create(fxfile);
			
			if(str_len(defines) > 1)
			{
				//map defines to new name
				var fileid = -1;
				var newid = 0;
				
				STRING *searchstr = str_create(fxfile);
				str_cat(searchstr, defines);
				
				var hndl = file_open_read(SDSF_FILEMAPPING_FILE);
				if(hndl > 0)
				{
					if(file_find(hndl, searchstr) > 0)
					{
						fileid = file_var_read(hndl);
					}else
					{
						file_seek(hndl, 0, 0);
						while(file_find(hndl, fxfile) > 0)
						{
							newid += 1;
						}
					}
					file_close(hndl);
				}
				
				if(fileid < 0)
				{
					hndl = file_open_append(SDSF_FILEMAPPING_FILE);
					file_str_write(hndl, searchstr);
					file_var_write(hndl, newid);
					fileid = newid;
					file_close(hndl);
				}
				
				str_trunc(outfile, 4);
				str_cat_num(outfile, "_%.0f.fxo", fileid);
				
				//set defines
				STRING *dns = str_create(defines);
				STRING *temp = str_create(defines);
				var pos = 0;
				while(pos = str_stri(dns, " "))
				{
					str_cpy(temp, dns);
					str_trunc(temp, str_len(temp)-pos+1);
					str_clip(dns, pos);
					
					str_cat(params, "/D ");
					str_cat(params, temp);
					str_cat(params, " ");
				}
				str_cat(params, "/D ");
				str_cat(params, dns);
				str_cat(params, " ");
				
				ptr_remove(dns);
				ptr_remove(temp);
			}
			
			//set output file
			str_cat(params, "/Fo ");
			str_cat(params, outfile);
			str_cat(params, " ");
			
			//set source file
			str_cat(params, fn);
			
			//compile
			exec_wait("slindev/utilities/fxc.exe", params);
			ptr_remove(params);
			
			ptr_remove(outfile);
		}
	}
	ptr_remove(fn);
}

/**
 * Create material.
 * Creates a new material with the shader from the given .fx (has to be .fxo if it is supposed to be precompiled) file.
 * @param fxfile name of the .fx file to load the shader from, with .fxo as extension if it is supposed to be precompiled.
 * @param defines a string with the defines to compile the shader with, seperated by spaces: "TESTDEF=1 ANOTHERDEF=1".
 * @return pointer to the new material
 */
MATERIAL *sdsf_mtl_create(STRING *fxfile, STRING *defines)
{
	MATERIAL *mat = mtl_create();
	
#ifdef SDSF_COMPILE
	sdsf_fx_compile(fxfile, defines);
#endif
	
	if(str_stri(fxfile, ".fxo"))
	{
#ifndef SDSF_IGNORE_COMPILED	
		//map defines to filename of the corresponding compiled file
		STRING *mappedfile = str_create(fxfile);
		if(str_len(defines) <= 1)
		{
			STRING *searchstr = str_create(fxfile);
			str_cat(searchstr, defines);	
			var hndl = file_open_read(SDSF_FILEMAPPING_FILE);
			if(hndl > 0)
			{
				if(file_find(hndl, searchstr) > 0)
				{
					str_trunc(mappedfile, 4);
					str_cat_num(mappedfile, "_%.0f.fxo", file_var_read(hndl));
				}
			}
		}
		
		if(file_exists(mappedfile))
		{
			effect_load(mat, mappedfile);
			return mat;
		}
#endif
		str_trunc(fxfile, 1);
	}
	
	if(str_len(defines) <= 1)
	{
		effect_load(mat, fxfile);
		return mat;
	}
	
	var hndl = file_open_read(fxfile);
	var filelen = file_length(hndl);
	char *buffer = (char *)sys_malloc(filelen*sizeof(char));
	file_str_readto(hndl, buffer, "", filelen);
	file_close(hndl);
	
	//construct defines string
	STRING *findefs = str_create("");
	STRING *dns = str_create(defines);
	STRING *temp = str_create(defines);
	var pos = 0;
	while(pos = str_stri(dns, " "))
	{
		str_cpy(temp, dns);
		str_trunc(temp, str_len(temp)-pos+1);
		str_clip(dns, pos);
		
		str_cat(findefs, "#define ");
		str_cat(findefs, temp);
		str_cat(findefs, "\n");
	}
	str_cat(findefs, "#define ");
	str_cat(findefs, dns);
	str_cat(findefs, "\n");
	while(pos = str_stri(findefs, "="))
	{
		str_setchr(findefs, pos, ' ');
	}
	
	//compile shader with defines
	STRING *shader = str_create(findefs);
	str_cat(shader, "\n");
	str_cat(shader, buffer);
	sys_free(buffer);
	effect_load(mat, shader);
	ptr_remove(shader);
	
	return mat;
}

/**
 * Create material.
 * Creates a new material with the shader from the given .fx (has to be .fxo if it is supposed to be precompiled) file.
 * @param fxfile name of the .fx file to load the shader from, with .fxo as extension if it is supposed to be precompiled.
 * @return pointer to the new material
 */
MATERIAL *sdsf_mtl_create(STRING *fxfile)
{
	return sdsf_mtl_create(fxfile, "");
}

#endif
